package com.streambase.contrib.erjang.impl;

import static org.junit.Assert.*;

import java.util.ArrayList;
import java.util.Date;
import java.util.List;

import org.junit.Before;
import org.junit.Test;

import com.streambase.sb.ByteArrayView;
import com.streambase.sb.CompleteDataType;
import com.streambase.sb.NullValueException;
import com.streambase.sb.Schema;
import com.streambase.sb.StreamBaseException;
import com.streambase.sb.Timestamp;
import com.streambase.sb.Tuple;
import com.streambase.sb.TupleException;

import erjang.EObject;
import erjang.ERT;

public class TypeMappingTest {
	private static Schema.Field blob = new Schema.Field("blob", CompleteDataType.forBlob());
	private static Schema.Field bool = new Schema.Field("bool", CompleteDataType.forBool());
	private static Schema.Field doub = new Schema.Field("doub", CompleteDataType.forDouble());
	private static Schema.Field intr = new Schema.Field("intr", CompleteDataType.forInt());
	private static Schema.Field lngr = new Schema.Field("lngr", CompleteDataType.forLong());
	private static Schema.Field str  = new Schema.Field("str", CompleteDataType.forString());
	private static Schema.Field ts  = new Schema.Field("ts", CompleteDataType.forTimestamp());

	private static Schema prim = 
		new Schema("PrimitiveTypeSchema", new Schema.Field[] {
			blob, bool, doub, intr, lngr, str, ts
	});
	
	private static Schema nested =
		new Schema("NestedTupleSchema", new Schema.Field[] {
			new Schema.Field("nested", CompleteDataType.forTuple(prim)),
		});

	private static Schema list =
		new Schema("ListOfTypesSchema", new Schema.Field[] {
			new Schema.Field("lblob", CompleteDataType.forList(blob.getCompleteDataType())),
			new Schema.Field("lbool", CompleteDataType.forList(bool.getCompleteDataType())),
			new Schema.Field("ldoub", CompleteDataType.forList(doub.getCompleteDataType())),
			new Schema.Field("lintr", CompleteDataType.forList(intr.getCompleteDataType())),
			new Schema.Field("llngr", CompleteDataType.forList(lngr.getCompleteDataType())),
			new Schema.Field("lstr", CompleteDataType.forList(str.getCompleteDataType())),
			new Schema.Field("lts", CompleteDataType.forList(ts.getCompleteDataType())),
			new Schema.Field("lnest", CompleteDataType.forList(CompleteDataType.forTuple(nested))),
		});
	
	
	@Before
	public void before() {
//		ErjangSession.register("erjang", "C:\\Program Files (x86)\\erl5.7.5", "5.7.5", new String[] { "C:\\workspaces\\sb70-spikes-erlang\\SbErjang\\erlang-src\\"} );
	}

	@Test
	public void testPrimitiveTypeMappings() throws StreamBaseException {
		Tuple primTuple = prim.createTuple();
		
		ByteArrayView binary = ByteArrayView.makeView(new byte[] { (byte)0xCA, (byte)0xFE, (byte)0xBA, (byte)0xBE });
		shook("blob", primTuple, (byte[])null, "dontcare");
		shook("blob", primTuple, ByteArrayView.makeView("horse"), "<<\"horse\">>");
		shook("blob", primTuple, binary, "<<202,254,186,190>>");
		shook("bool", primTuple, (Boolean)null, "dontcare");
		shook("bool", primTuple, (Boolean)false, "false");
		shook("bool", primTuple, (Boolean)true, "true");
		shook("doub", primTuple, (Double)null, "dontcare");
		shook("doub", primTuple, Double.MIN_VALUE, "" + Double.MIN_VALUE);
		shook("doub", primTuple, Double.MAX_VALUE, "" + Double.MAX_VALUE);
		shook("intr", primTuple, (Integer)null, "dontcare");
		shook("intr", primTuple, Integer.MIN_VALUE, "" + Integer.MIN_VALUE);
		shook("intr", primTuple, Integer.MAX_VALUE, "" + Integer.MAX_VALUE);
		shook("lngr", primTuple, (Long)null, "dontcare");
		shook("lngr", primTuple, Long.MIN_VALUE, "" + Long.MIN_VALUE);
		shook("lngr", primTuple, Long.MAX_VALUE, "" + Long.MAX_VALUE);
		shook("str", primTuple, (String)null, "dontcare");
		shook("str", primTuple, "-9223372036854775808", "\"-9223372036854775808\"");
//		shook("str", primTuple, "\r\n\f -9223372036854775808", "\"\\r\\n\\x0C -9223372036854775808\"");
		shook("ts", primTuple, null, "dontcare");
		shook("ts", primTuple, Timestamp.setValue(101L), "{interval,50}");
		shook("ts", primTuple, Timestamp.fromString("2011-03-30 12:43:03.3333"), "{timestamp,{1301,485383,333000}}");
	}
	
	@Test
	public void testTupleTypeMappings() throws NullValueException, TupleException {
		Tuple nestedTuple = nested.createTuple();
		Tuple primTuple = prim.createTuple();
		shook("nested", nestedTuple, (Tuple)null, "dontcare");
		nestedTuple.setTuple(0, primTuple);
		shook("nested", nestedTuple, primTuple, "{[],[],[],[],[],[],[]}");
		primTuple.setField("blob","horse");
		shook("nested", nestedTuple, primTuple, "{<<\"horse\">>,[],[],[],[],[],[]}");
		primTuple.setBoolean("bool", true);
		shook("nested", nestedTuple, primTuple, "{<<\"horse\">>,true,[],[],[],[],[]}");
		primTuple.setDouble("doub", Double.NaN);
		shook("nested", nestedTuple, primTuple, "{<<\"horse\">>,true,NaN,[],[],[],[]}");
		primTuple.setInt("intr", 42);
		shook("nested", nestedTuple, primTuple, "{<<\"horse\">>,true,NaN,42,[],[],[]}");
		primTuple.setLong("lngr", 666L);
		shook("nested", nestedTuple, primTuple, "{<<\"horse\">>,true,NaN,42,666,[],[]}");
		primTuple.setString("str", "foobar");
		shook("nested", nestedTuple, primTuple, "{<<\"horse\">>,true,NaN,42,666,\"foobar\",[]}");
		primTuple.setTimestamp("ts", new Timestamp(new Date(123456789)));
		shook("nested", nestedTuple, primTuple, "{<<\"horse\">>,true,NaN,42,666,\"foobar\",{timestamp,{0,123456,789000}}}");
	}

	@Test
	public void testListTypeMappings() throws NullValueException, TupleException {
		Tuple listTuple = list.createTuple();
		List<Object> elms = new ArrayList<Object>();
		elms.add(ByteArrayView.makeView(new byte[] { (byte)0xF0, (byte)0x0B, (byte)0xA5 }));
		elms.add(ByteArrayView.makeView(new byte[] { (byte)0xF0, (byte)0x0B, (byte)0xA6 }));
		elms.add(ByteArrayView.makeView(new byte[] { (byte)0xF0, (byte)0x0B, (byte)0xA7 }));
		elms.add(ByteArrayView.makeView(new byte[] { (byte)0xF0, (byte)0x0B, (byte)0xA8 }));
		listTuple.setList("lblob", elms);
		shook("lblob", listTuple, elms, "[<<240,11,165>>,<<240,11,166>>,<<240,11,167>>,<<240,11,168>>]");
		elms.clear(); elms.add(Boolean.FALSE); elms.add(Boolean.TRUE);
		listTuple.setList("lbool", elms);
		shook("lbool", listTuple, elms, "[false,true]");
		elms.clear(); elms.add(Double.NaN); elms.add(Double.NEGATIVE_INFINITY);
		listTuple.setList("ldoub", elms);
		shook("ldoub", listTuple, elms, "[NaN,-Infinity]");
		elms.clear(); elms.add(Integer.MAX_VALUE); elms.add(Integer.MIN_VALUE);
		listTuple.setList("lintr", elms);
		shook("lintr", listTuple, elms, "[2147483647,-2147483648]");
		elms.clear(); elms.add(Long.MIN_VALUE); elms.add(Long.MAX_VALUE);
		listTuple.setList("llngr", elms);
		shook("llngr", listTuple, elms, "[-9223372036854775808,9223372036854775807]");
		elms.clear(); elms.add("foo"); elms.add("bar");
		listTuple.setList("lstr", elms);
		shook("lstr", listTuple, elms, "[\"foo\",\"bar\"]");
		elms.clear(); elms.add(Timestamp.msecs(Timestamp.INTERVAL, 42)); elms.add(Timestamp.msecs(Timestamp.TIMESTAMP, 0));
		listTuple.setList("lts", elms);
		shook("lts", listTuple, elms, "[{interval,42},{timestamp,{0,0,0}}]");
	}
	
	// An S-hook is a butchers device for hanging meat on a rail. In testing an S-hook is a means to
	// transform from one type system to another and back again. If the to transform meets expectations
	// and the inverse transforms returns the type to its original lexical/value space without loss then
	// the test has succeeded. An S-hook is a convenient way to ensure type transformations are symmetric.
	//
	public static <T> void shook(String field, Tuple x, T sb, String erl) throws NullValueException, TupleException {
		if (sb == null) {
			x.setNull(field);
			EObject s2e = TypeMapping.sbToErjang(x, x.getSchema().getField(field));
			assertEquals(ERT.NIL, s2e);
			assertNull(TypeMapping.erjangToTuple(x.getSchema().createTuple(),x.getSchema().getField(field),s2e).getField(field));
		} else {
			x.setField(field, sb);
			EObject s2e = TypeMapping.sbToErjang(x, x.getSchema().getField(field));
			assertEquals(erl, s2e.toString());
			if ("str".equals(field)) {
				assertEquals(""+sb,TypeMapping.erjangToTuple(x.getSchema().createTuple(), x.getSchema().getField(field),s2e).getField(field));
			} else {
				Tuple t = TypeMapping.erjangToTuple(x.getSchema().createTuple(), x.getSchema().getField(field),s2e);
				if (sb instanceof Tuple) {
					assertEquals(((Tuple)sb),t.getTuple(field));
				} else {
					assertEquals(sb, t.getField(field));
				}
			}
		}
	}
}
